#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "fly.h"

#if 0
#define DEBUG_UpdateViewer
#endif

#define TIME_DELAY  500

#define X 0
#define Y 1
#define Z 2

#define Umin 0
#define Umax 1
#define Vmin 2
#define Vmax 3


static status = 0;
struct configviewer {
   /* Viewer geometry */
   char  *geometry;

   /* Control parameter initialized, when the viewer is set up */
   /* Motion modifier range */

   double  MaxSpeed;
   double  MinSpeed;
   double  TurnLeftLimit;
   double  TurnRightLimit;
   double  MaxClimb;
   double  MinClimb;

   /* View modifier */
   double  ViewLeftLimit;
   double  ViewRightLimit;
   double  ViewDownLimit;

   /* Inital viewing parameter */
   char   *StartPosition;
   double  InitialHeading;
};

typedef struct configviewer ConfigViewer;

static ConfigViewer  configviewer;
static Tk_ConfigSpec configSpecs[] = {
   {TK_CONFIG_STRING, "-geometry", "geometry", "Geometry", "640x400",
       Tk_Offset(ConfigViewer, geometry), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-maxspeed", "maxspeed", "MaxSpeed", "100",
       Tk_Offset(ConfigViewer, MaxSpeed), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-minspeed", "minspeed", "MinSpeed", "0",
       Tk_Offset(ConfigViewer, MinSpeed), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-turnleftlimit", "turnleftlimit", "TurnLeftLimit", "-90",
       Tk_Offset(ConfigViewer, TurnLeftLimit), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-turnrightlimit", "turnrightlimit", "TurnRightLimit",
       "90", Tk_Offset(ConfigViewer, TurnRightLimit), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-maxclimb", "maxclimb", "MaxClimb", "-15",
       Tk_Offset(ConfigViewer, MaxClimb), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-minclimb", "minclimb", "MinClimb", "15",
       Tk_Offset(ConfigViewer, MinClimb), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-viewleftlimit", "viewleftlimit", "ViewLeftLimit", "-90",
       Tk_Offset(ConfigViewer, ViewLeftLimit), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-viewrightlimit", "viewrightlimit", "ViewRightLimit",
       "-90", Tk_Offset(ConfigViewer, ViewRightLimit), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-viewdownlimit", "viewdownlimit", "ViewDownLimit", "-90",
       Tk_Offset(ConfigViewer, ViewDownLimit), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_STRING, "-startposition", "startposition", "StartPosition",
       "(0.0, 0.0, 0.0)", Tk_Offset(ConfigViewer, StartPosition),
       TK_CONFIG_NULL_OK, (Tk_CustomOption *) NULL},
   {TK_CONFIG_DOUBLE, "-initialheading", "initialheading", "InitialHeading",
       "0", Tk_Offset(ConfigViewer, InitialHeading), TK_CONFIG_NULL_OK,
       (Tk_CustomOption *) NULL},
   {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
   (char *) NULL, TK_CONFIG_NULL_OK, (Tk_CustomOption *) NULL}
};

int             ConfigureViewer(Tcl_Interp *interp, Viewer *viewer,
                   int argc, char *argv[])
{
   XShmSegmentInfo *shminfo = &(viewer->pixmapshminfo);
   XImage          *image;
   int              width;
   int              height;
   XGCValues        gcValues;
   XRectangle       rect;
   int              Major;
   int              Minor;
   Bool             yes_pixmap;

   if (!XShmQueryVersion(viewer->display, &Major, &Minor, &yes_pixmap)) {
      Tcl_AppendResult(viewer->interp, "Shared memory not supported",
                       (char *) NULL);
      return(TCL_ERROR);
   }
#if 0
   if (yes_pixmap == True)
      printf("yes Shared pixmap\n");
   else
      printf("no Shared pixmap\n");
#endif
   if (Tk_ConfigureWidget(interp, viewer->tkwin, configSpecs, argc, argv,
                          (char *) &configviewer, 0) == TCL_OK)
   {

      sscanf(configviewer.geometry, "%dx%d", &width, &height);
      printf("(width, height) = (%d, %d)\n", width, height);
      Tk_GeometryRequest(viewer->tkwin, width, height);

      viewer->width = width;
      viewer->height = height;

      shminfo = &(viewer->imageshminfo);
      image = XShmCreateImage(viewer->display, viewer->visual,
                              (unsigned int) viewer->depth,
                              ZPixmap, (char *) NULL, shminfo,
                              (unsigned int) width, (unsigned int) height);
      if (image == NULL)
      {
         Tcl_AppendResult(viewer->interp,
                          "allocation failure in XShmCreateImage",
                          (char *) NULL);
         return(TCL_ERROR);
      }
      viewer->image = image;

      printf("image->bits_per_pixel = %d\n", image->bits_per_pixel);
      shminfo->shmid = shmget(IPC_PRIVATE, image->bytes_per_line
                                           * image->height, IPC_CREAT | 0777);
      if (shminfo->shmid < 0)
      {
         Tcl_AppendResult(viewer->interp, "Shared memory error in shmget",
                          (char *) NULL);
         return(TCL_ERROR);
      }
 
      shminfo->shmaddr = (char *) shmat(shminfo->shmid, (char *) 0, 0);

      if (shminfo->shmaddr == ((char *) -1)) {
         Tcl_AppendResult(viewer->interp, "Shared memory error in shmaddr",
                          (char *) NULL);
         return(TCL_ERROR);
      }

      shminfo->readOnly = False;

      XShmAttach(viewer->display, shminfo);

      viewer->image->data = shminfo->shmaddr;

      shminfo = &(viewer->pixmapshminfo);
      shminfo->shmid = shmget(IPC_PRIVATE, image->bytes_per_line
                                           * image->height, IPC_CREAT | 0777);
      if (shminfo->shmid < 0)
      {
         Tcl_AppendResult(viewer->interp, "Shared memory error in shmget",
                          (char *) NULL);
         return(TCL_ERROR);
      }
 
      shminfo->shmaddr = (char *) shmat(shminfo->shmid, (char *) 0, 0);

      if (shminfo->shmaddr == ((char *) -1)) {
         Tcl_AppendResult(viewer->interp, "Shared memory error in shmaddr",
                          (char *) NULL);
         return(TCL_ERROR);
      }

      shminfo->readOnly = False;

      XShmAttach(viewer->display, shminfo);

      viewer->root = RootWindow(viewer->display,
                                Tk_ScreenNumber(viewer->tkwin));

      viewer->pixelmap = XShmCreatePixmap(viewer->display, viewer->root,
                                        shminfo->shmaddr, shminfo,
                                        width, height, viewer->depth);
      rect.x = 0;
      rect.y = 0;
      rect.width = width;       /* these are unsigned shorts */
      rect.height = height;

      viewer->imageregion = XCreateRegion();
      XUnionRectWithRegion(&rect, viewer->imageregion, viewer->imageregion);

      viewer->white = Tk_GetColor(viewer->interp, viewer->tkwin, None, "white");
      viewer->black = Tk_GetColor(viewer->interp, viewer->tkwin, None, "black");

      gcValues.foreground = (viewer->white != NULL) ? viewer->white->pixel :
         WhitePixelOfScreen(Tk_Screen(viewer->tkwin));
      gcValues.background = (viewer->black != NULL) ? viewer->black->pixel :
         BlackPixelOfScreen(Tk_Screen(viewer->tkwin));
      gcValues.graphics_exposures = False;

      viewer->blankGC = Tk_GetGC(viewer->tkwin, GCForeground | GCBackground,
                                 &gcValues);

      printf("Before Tk_SetWindowBackground\n");
      Tk_SetWindowBackground(viewer->tkwin,
                             BlackPixelOfScreen(Tk_Screen(viewer->tkwin)));

      viewer->control->MaxSpeed = configviewer.MaxSpeed;
      viewer->control->MinSpeed = configviewer.MinSpeed;

      viewer->control->TurnLeftLimit = configviewer.TurnLeftLimit;
      viewer->control->TurnRightLimit = configviewer.TurnRightLimit;

      viewer->control->MaxClimb = configviewer.MaxClimb;
      viewer->control->MinClimb = configviewer.MinClimb;

      viewer->control->ViewLeftLimit = configviewer.ViewLeftLimit;
      viewer->control->ViewRightLimit = configviewer.ViewRightLimit;
      viewer->control->ViewDownLimit = configviewer.ViewDownLimit;

      return(TCL_OK);
   }
   else
      return(TCL_ERROR);
}

void            UpdateViewer(ClientData clientdata)
{
   extern int  newimage;
   Viewer     *viewer = (Viewer *) clientdata;
   Display    *display = viewer->display;
   GC          blankGC = viewer->blankGC;
   int         width = viewer->width;
   int         height = viewer->height;

#if 0
   printf("UpdateViewer(): newimage = %s\n", newimage == TRUE ? "TRUE" :
          "FALSE");
#endif

   if ((viewer->tkwin != NULL) && Tk_IsMapped(viewer->tkwin))
   {
#ifdef DEBUG_UpdateViewer
      printf("UpdateViewer() -- before XShmPutImage\n");
#endif
      XShmPutImage(display, viewer->pixelmap, blankGC, viewer->image,
                   0, 0, 0, 0, width, height, True);
      XSetRegion(display, blankGC, viewer->imageregion);
      XCopyArea(display, viewer->pixelmap, Tk_WindowId(viewer->tkwin),
                blankGC, 0, 0, width, height, 0, 0);
      kill(viewer->pid, SIGUSR2);
   }
#if 0
   viewer->timer = Tk_CreateTimerHandler(TIME_DELAY, UpdateViewer,
                                         (ClientData) viewer);
#endif
}

void            ExposedViewer(ClientData clientdata)
{
   static int      first = 0;
   Viewer         *viewer = (Viewer *) clientdata;
   Display        *display = viewer->display;
   GC              blankGC = viewer->blankGC;
   int             width = viewer->width;
   int             height = viewer->height;

   viewer->status &= ~SET_ExposedViewer;

   if ((viewer->tkwin == NULL) || !(Tk_IsMapped(viewer->tkwin)))
      return;

#if 0
   printf("ExposedViewer() -- started\n");
#endif
   if (first == 0)
   {                            /* First time called, so initialize first */
#if 0
      XSetForeground(display, blankGC, viewer->black->pixel);
#endif
      XSetForeground(display, blankGC, 0xffff00);
      XFillRectangle(display, viewer->pixelmap, blankGC, 0, 0, width,
                     height);
      kill(viewer->pid, SIGALRM);

      first = 1;
   }
   XSetRegion(display, blankGC, viewer->imageregion);
   XCopyArea(display, viewer->pixelmap, Tk_WindowId(viewer->tkwin),
             blankGC, 0, 0, width, height, 0, 0);
}

void            DestroyViewer(ClientData clientdata)
{
   Viewer         *viewer = (Viewer *) clientdata;

   Tk_FreeOptions(configSpecs, (char *) &configviewer, viewer->display, 0);

   Tk_FreeGC(viewer->display, viewer->blankGC);

   XShmDetach(viewer->display, &(viewer->pixmapshminfo));
   shmdt(viewer->pixmapshminfo.shmaddr);
   shmctl(viewer->pixmapshminfo.shmid, IPC_RMID, (struct shmid_ds *) 0);
   XFreePixmap(viewer->display, viewer->pixelmap);

   XShmDetach(viewer->display, &(viewer->imageshminfo));
   shmdt(viewer->imageshminfo.shmaddr);
   shmctl(viewer->imageshminfo.shmid, IPC_RMID, (struct shmid_ds *) 0);
   XDestroyImage(viewer->image);

   shmdt(viewer->lineshminfo.shmaddr);
   shmctl(viewer->lineshminfo.shmid, IPC_RMID, (struct shmid_ds *) 0);

   viewer->display = NULL;

   XFree(viewer->white);
   XFree(viewer->black);

   if (viewer->depth > 8)
   {
      cfree(viewer->colors);
      cfree(viewer->pixels);
   }

   if (viewer->datafilename != NULL)
      free(viewer->datafilename);

   if (viewer->control != NULL)
      free(viewer->control);
#if 0
   if (viewer->object != NULL)
      free(viewer->object);
#endif
   free(viewer);
}

void            windowEvent(ClientData clientdata, XEvent * eventPtr)
{
   Viewer         *viewer = (Viewer *) clientdata;

   if (eventPtr->type == Expose)
   {
      if ((eventPtr->xexpose.count == 0))
      {
         viewer->status |= SET_ExposedViewer;
         Tk_DoWhenIdle(ExposedViewer, clientdata);
      }
   }
   else if (eventPtr->type == DestroyNotify)
   {
      kill(viewer->pid, SIGKILL);
      Tcl_DeleteCommand(viewer->interp, Tk_PathName(viewer->tkwin));
      viewer->tkwin = NULL;
      if (viewer->status & SET_ExposedViewer)
      {
         Tk_CancelIdleCall(ExposedViewer, clientdata);
         viewer->status &= ~SET_ExposedViewer;
      }
#if 0
      if (viewer->status & SET_Timer)
      {
         Tk_DeleteTimerHandler(viewer->timer);
         viewer->status &= ~SET_Timer;
      }
#endif
      Tk_EventuallyFree(clientdata, DestroyViewer);
   }
}

int             viewercontrol(ClientData clientData, Tcl_Interp *interp,
                              int argc, char *argv[])
{
   Viewer         *viewer = (Viewer *) clientData;
   Control        *control = viewer->control;
   int             command_len;

   if (argc < 3)
   {
      Tcl_AppendResult(interp,
                       "wrong # args: should be \"pathName command parameter\"",
                       (char *) NULL);
      return (TCL_ERROR);
   }

   command_len = strlen(argv[1]);
   if (strncmp(argv[1], "speed", command_len) == 0)
   {
      /*
       * Parse the speed option.
       */
      if (Tcl_GetDouble(interp, argv[2], &(control->Speed)) == TCL_OK)
      {
         if ((control->Speed < control->MinSpeed)
             || (control->Speed > control->MaxSpeed))
         {
            Tcl_AppendResult(interp, "Speed is not within range",
               (char *) NULL);
            return TCL_ERROR;
         }
         printf("Control Speed %g\n", control->Speed);
      }
      else
         return TCL_ERROR;
   }
   else if (strncmp(argv[1], "turnangle", command_len) == 0)
   {
      /*
       * Parse the turnangle option.
       */
      if (Tcl_GetDouble(interp, argv[2], &(control->TurnAngle)) == TCL_OK)
      {
         if ((control->TurnAngle < control->TurnLeftLimit)
             || (control->TurnAngle > control->TurnRightLimit))
         {
            Tcl_AppendResult(interp, "Turn is not within range",
               (char *) NULL);
            return TCL_ERROR;
         }
         printf("Control TurnAngle %g\n", control->TurnAngle);
      }
      else
         return TCL_ERROR;
   }
   else if (strncmp(argv[1], "climb", command_len) == 0)
   {
      /*
       * Parse the climb option.
       */
      if (Tcl_GetDouble(interp, argv[2], &(control->Climb)) == TCL_OK)
      {
         if ((control->Climb < control->MinClimb)
            || (control->Climb > control->MaxClimb))
         {
            Tcl_AppendResult(interp, "Climb is not within range",
               (char *) NULL);
            return TCL_ERROR;
         }
         printf("Control Climb %g\n", control->Climb);
      }
      else
         return TCL_ERROR;
   }
   else if (strncmp(argv[1], "viewsideangle", command_len) == 0)
   {
      /*
       * Parse the viewsideangle option.
       */
      if (Tcl_GetDouble(interp, argv[2], &(control->ViewSideAngle)) == TCL_OK)
      {
         if ((control->ViewSideAngle < control->ViewLeftLimit)
            || (control->ViewSideAngle > control->ViewRightLimit))
         {
            Tcl_AppendResult(interp, "Side is not within range",
               (char *) NULL);
            return TCL_ERROR;
         }
         printf("Control ViewSideAngle %g\n", control->ViewSideAngle);
      }
      else
         return TCL_ERROR;
   }
   else if (strncmp(argv[1], "viewdownangle", command_len) == 0)
   {
      /*
       * Parse the -viewdownangle option.
       */
      if (Tcl_GetDouble(interp, argv[2], &(control->ViewDownAngle)) == TCL_OK)
      {
         if ((control->ViewDownAngle < control->ViewDownLimit)
            || (control->ViewDownAngle > 0.0))
         {
            Tcl_AppendResult(interp, "Down is not within range",
               (char *) NULL);
            return TCL_ERROR;
         }
         printf("Control ViewDownAngle %g\n", control->ViewDownAngle);
      }
      else
         return TCL_ERROR;
   }
   else if (strncmp(argv[1], "info", command_len) == 0)
   {
      printf("not impremented yet\n");
   }
   else
   {
      Tcl_AppendResult(interp, "Unknow Flag", argv[1], (char *) NULL);
      return TCL_ERROR;
   }

   return (TCL_OK);
}
#if 0
void get_new_visual_info(ParentWindow * parentwindow,
                         XVisualInfo *new_visual_info);
void get_new_colormap(parentwindow, &new_visual_info, &colormap);
#endif
int             viewercmd(ClientData clientData, Tcl_Interp * interp,
                          int argc, char *argv[])
{
   ParentWindow    *parentwindow = (ParentWindow *) clientData;
   Viewer          *viewer;
   XShmSegmentInfo *shminfo;
   Tk_Window        window;
   FILE            *datafile;

   int            *Depth_List;
   int             Depth_count;
   int             Max_Depth;

   XVisualInfo     visual_info;
   XVisualInfo    *visual_list;
   int             n_visual;

   int             i;

   Colormap        colormap;
   char            color_name[256];
   unsigned long   plane_masks[1];
   int             ncolor;
   XColor         *colors;
   unsigned long  *pixels;

   if (argc < 3)
   {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
         " pathName filename ?options?\"", (char *) NULL);
      return (TCL_ERROR);
   }

   if ((datafile = fopen(argv[2], "r")) == NULL)
   {
      Tcl_AppendResult(interp, "could not open datafile file", argv[2],
                       (char *) NULL);
      return (TCL_ERROR);
   }
   fclose(datafile);

   visual_info.screen = Tk_ScreenNumber(parentwindow->tkwin);
   if ((Depth_List = XListDepths(parentwindow->display, visual_info.screen,
                                 &Depth_count)) != NULL)
   {
      Max_Depth = Depth_List[Depth_count - 1];
      for (i = Depth_count - 2; i >= 0; i--)
         if (Max_Depth < Depth_List[i])
            Max_Depth = Depth_List[i];
   }
   else
   {
      fprintf(stderr, "Cannot get depth list\n");
      exit(1);
   }
   
   XFree((char *) Depth_List);

   printf("Screen depth = %d\n", Max_Depth);

   if (Max_Depth > 8)
   {
      visual_info.depth = Max_Depth;
      if ((visual_list = XGetVisualInfo(parentwindow->display,
                                        VisualScreenMask | VisualDepthMask,
                                        &visual_info, &n_visual)) == NULL)
      {
         fprintf(stderr,"no visual information available\n");
         exit(1);
      }

      fprintf(stderr, "Number of visual = %d\n", n_visual);

      /*
       * Remember various properties of this visual.
       */
      for (i = 0; i < n_visual; i++)
         switch (visual_list[i].class)
         {
            case DirectColor:
               visual_info = visual_list[i];
               fprintf(stderr, "Visual = DirectColor\n");
               break;
            case TrueColor:
               fprintf(stderr, "Visual = TrueColor\n");
               break;
            case PseudoColor:
               fprintf(stderr, "Visual = PseudoColor\n");
               break;
            case StaticColor:
               fprintf(stderr, "Visual = StaticColor\n");
               break;
            case GrayScale:
               fprintf(stderr, "Visual = GrayScale\n");
               break;
            case StaticGray:
               fprintf(stderr, "Visual = StaticGray\n");
               break;
         }

      XFree((char *) visual_list);

      colormap = XCreateColormap(parentwindow->display,
                    RootWindowOfScreen(Tk_Screen(parentwindow->tkwin)),
                                 visual_info.visual, AllocNone);

      ncolor = visual_info.visual->map_entries;
      colors = (XColor *) ckalloc(ncolor * sizeof(XColor));
      pixels = (unsigned long *) ckalloc(ncolor * sizeof(unsigned long));

      if (!XAllocColorCells(parentwindow->display, colormap,
                            False, plane_masks, 0, pixels, ncolor))
      {
         printf("Allocate color fail\n");
         exit(1);
      }

      printf("ncolor = %d\n", ncolor);
      for (i = 0; i < ncolor; i++)
      {
         sprintf(color_name, "#%06lx", pixels[i]);
#ifdef DEBUG_color
         printf("color_name[%d] = (%s)\n", i, color_name);
#endif
         if (XParseColor(parentwindow->display, colormap, color_name,
                         &colors[i]))
         {
            colors[i].pixel = pixels[i];
            colors[i].flags = DoRed | DoGreen | DoBlue;
#ifdef DEBUG_color
            printf("(Red, Green, Blue) = (%hd, %hd, %hd)\n",
                   colors[i].red, colors[i].green, colors[i].blue);
#endif
         }
         else
         {
            printf("XParseColor failed\n");
            exit(1);
         }
      }
      XStoreColors(parentwindow->display, colormap, colors, ncolor);
   }
   else
   {
      visual_info.visualid = XVisualIDFromVisual(parentwindow->visual);
      if ((visual_list = XGetVisualInfo(parentwindow->display,
                                        VisualScreenMask | VisualIDMask,
                                        &visual_info, &n_visual)) == NULL)
      {
         fprintf(stderr,"no visual information available\n");
         exit(1);
      }
      visual_info = visual_list[0];
      XFree((char *) visual_list);
   }

   if ((window = Tk_CreateWindowFromPath(interp, parentwindow->tkwin,
                                         argv[1], (char *) NULL)) != NULL)
   {
      if (Max_Depth > 8)
      {
         if (!Tk_SetWindowVisual(window, visual_info.visual,
                                 visual_info.depth, colormap))
         {
            Tcl_AppendResult(interp, "Did not change visual.", (char *) NULL);
            Tk_DestroyWindow(window);
            return(TCL_ERROR);
         }
         Tk_SetWindowColormap(window, colormap);
      }

      Tk_SetClass(window, "Viewer");

      if (((viewer = (Viewer *) malloc(sizeof(Viewer))) != NULL)
          && ((viewer->control = (Control *) malloc(sizeof(Control))) != NULL))
      {
         status = 1;
         viewer->parentwindow = parentwindow;
         viewer->tkwin = window;
         viewer->interp = parentwindow->interp;
         viewer->display = Tk_Display(window);
         viewer->status = SET_none;
         viewer->datafilename = (char *) strdup(argv[2]);

         viewer->visual_info = visual_info;
         viewer->depth = visual_info.depth;
         viewer->visual = visual_info.visual;

         if (Max_Depth > 8)
         {
            viewer->colormap = colormap;
            viewer->ncolor = ncolor;
            viewer->colors = colors;
            viewer->pixels = pixels;
         }
         else
         {
            viewer->colormap = None;
            viewer->ncolor = 0;
            viewer->colors = NULL;
            viewer->pixels = NULL;
         }

#if 0
         viewer->object = NULL;
#endif
         shminfo = &(viewer->lineshminfo);

         shminfo->shmid = shmget(IPC_PRIVATE, sizeof(LineData),
                                 IPC_CREAT | 0777);
         if (shminfo->shmid < 0)
         {
            Tcl_AppendResult(viewer->interp, "Shared memory error in shmget",
                          (char *) NULL);
            return(TCL_ERROR);
         }

         shminfo->shmaddr = (char *) shmat(shminfo->shmid, 0, 0);
         viewer->linedata = (LineData *) shminfo->shmaddr;

         if (shminfo->shmaddr == ((char *) -1)) {
            Tcl_AppendResult(viewer->interp, "Shared memory error in shmaddr",
                             (char *) NULL);
            return(TCL_ERROR);
         }

         shminfo->readOnly = False;

         if (ConfigureViewer(interp, viewer, argc - 3, argv + 3) == TCL_OK)
         {
            printf("Before Tk_CreateEventHandler\n");
            Tk_CreateEventHandler(window, ExposureMask | StructureNotifyMask,
               windowEvent, (ClientData) viewer);

            printf("Before Tcl_CreateCommand\n");
            Tcl_CreateCommand(interp, Tk_PathName(viewer->tkwin),
               (Tcl_CmdProc *) viewercontrol, (ClientData *) viewer,
               (Tcl_CmdDeleteProc *) NULL);
            interp->result = Tk_PathName(window);

            if (init_comm(&(viewer->comm)) == TCL_OK)
            {
               fprintf(stderr, "before fork()\n");
               viewer->ppid = getpid();
               switch (viewer->pid = fork())
               {
                  case 0:             /* In child */
                     close (ConnectionNumber(viewer->display));
                     fprintf(stderr, "Child: Child process started pid = %d\n",
                             getpid());
                     init_child_comm(&(viewer->comm));
                     init_child_message_handler(viewer);
                     child_main(viewer);
                     break;           /* Not reached */

                  case -1:            /* Error */
                     fprintf(stderr, "Could not start child process\n");
                     break;

                  default:            /* Parent */
                     fprintf(stderr, "Parent: Child process started pid = %d\n",
                             getpid());
                     init_parent_comm(&(viewer->comm));
                     init_parent_message_handler(viewer);
                     break;
               }
            }
#if 0
            viewer->status |= SET_Timer;
            viewer->timer = Tk_CreateTimerHandler(TIME_DELAY, UpdateViewer,
                                                  (ClientData) viewer);
#endif
            return(TCL_OK);
         }
         else
         {
            free(viewer->control);
            free(viewer);
            Tcl_AppendResult(interp, "Can not configure terrain viewer.",
                             (char *) NULL);
            Tk_DestroyWindow(window);
            return(TCL_ERROR);
         }
      }
      else
      {
         if (viewer != NULL)
            free(viewer);
         Tcl_AppendResult(interp, "Out of memory!", (char *) NULL);
         Tk_DestroyWindow(window);
         return(TCL_ERROR);
      }
   }
   else
   {
      Tcl_AppendResult(interp, "Can not create window.", (char *) NULL);
      return(TCL_ERROR);
   }
}
